# -*- coding: utf-8 -*-
"""Top-level package for pybeehive."""
from .core import Event, Listener, Streamer
from .hive import Hive

__author__ = """Djordje Pepic"""
__email__ = 'djordje.m.pepic@gmail.com'
__version__ = '0.1.1'

__all__ = [
    'Event', 'Listener', 'Streamer',
    'Hive'
]



from multiprocessing import Event as _Event
from queue import Queue, Empty
from threading import Thread, Event
import time
import _thread
import pytest
import pybeehive


def run_kill_hive(hive, **kwargs):
    wait = kwargs.pop('wait', 0.03)

    def kill_after():
        time.sleep(wait)
        hive.kill()

    runner = Thread(target=kill_after)
    runner.start()
    hive.run(**kwargs)


def test_run(hive, bee_factory):
    listener = bee_factory.create('listener')
    streamer = bee_factory.create('streamer')
    hive.add(listener)
    hive.add(streamer)
    # Run once in debug mode to make sure,
    # but this should never cause errors
    run_kill_hive(hive, debug=True)
    assert len(listener.calls) > 0, 'Run did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Run did not wrap non event data'


def test_decorated_listener(hive):
    calls = []

    @hive.listener
    def on_event(event):
        calls.append(event)
    assert len(hive.listeners) == 1, 'Did not register listener to hive'
    hive.submit_event(pybeehive.Event('test'))
    run_kill_hive(hive)
    assert len(calls) == 1, 'Listener did not execute on_event'
    assert calls[0].data == 'test', 'Listener did not return correct data'


def test_decorated_listener_chained(hive):
    chained_calls = []

    @hive.listener
    def listener1(event):
        return event

    @hive.listener(chain='listener1')
    def listener2(event):
        chained_calls.append(event)

    assert len(hive.listeners) == 1, 'Added chained listener to root list of listeners'
    assert len(hive.listeners._listeners['listener1'][0].chained_bees) == 1,\
        'Did not correctly chain listeners'
    hive.submit_event(pybeehive.Event('test'))
    run_kill_hive(hive)
    assert len(chained_calls) == 1, 'Chained listener did not execute on_event'


def test_decorated_listener_chained_multi(hive):
    chained_calls = []

    @hive.listener(filters=['topic1'])
    def listener1(event):
        return event

    @hive.listener(filters=['topic2'])
    def listener2(event):
        return event

    @hive.listener(chain=['listener1', 'listener2'])
    def listener3(event):
        chained_calls.append(event)

    assert len(hive.listeners) == 2, 'Added chained listener to root list of listeners'
    assert len(hive.listeners._listeners['listener1'][0].chained_bees) == 1,\
        'Did not correctly chain listeners'
    assert len(hive.listeners._listeners['listener2'][0].chained_bees) == 1,\
        'Did not correctly chain listeners'
    hive.submit_event(pybeehive.Event('test1', topic='topic1'))
    hive.submit_event(pybeehive.Event('test2', topic='topic2'))
    run_kill_hive(hive)
    assert len(chained_calls) == 2, 'Chained listener did not execute on_event from both chains'


def test_decorated_listener_with_filters(hive):
    all_calls = []
    filtered_calls = []

    @hive.listener
    def listener1(event):
        all_calls.append(event)

    @hive.listener(filters='topic1')
    def listener2(event):
        filtered_calls.append(event)

    hive.submit_event(pybeehive.Event('test'))
    hive.submit_event(pybeehive.Event('test1', topic='topic1'))
    run_kill_hive(hive)
    assert len(all_calls) == 2, 'Unfiltered listener did not execute all on_event calls'
    assert len(filtered_calls) == 1, 'Filtered listener did executed on_event when it shouldnt'


def test_decorated_listener_chained_with_filters(hive):
    topic1_calls = []
    topic2_calls = []

    @hive.listener
    def listener1(event):
        return event

    @hive.listener(chain=['listener1'], filters=['topic1'])
    def listener2(event):
        event.data = 'listener2'
        topic1_calls.append(event)

    @hive.listener(chain=['listener1'], filters=['topic2'])
    def listener3(event):
        event.data = 'listener3'
        topic2_calls.append(event)

    hive.submit_event(pybeehive.Event('test1', topic='topic1'))
    hive.submit_event(pybeehive.Event('test2', topic='topic2'))
    run_kill_hive(hive)
    assert len(topic1_calls) == 1, 'Chained filtered listener did not execute on_event'
    assert len(topic2_calls) == 1, 'Chained filtered listener did not execute on_event'
    assert topic1_calls[0].data == 'listener2', 'Chained filtered listener did not execute on_event'
    assert topic2_calls[0].data == 'listener3', 'Chained filtered listener did not execute on_event'


def test_decorated_streamer(hive, bee_factory):
    listener = bee_factory.create('listener')
    hive.add(listener)

    @hive.streamer
    def stream():
        for i in range(5):
            yield i
    assert len(hive.streamers) == 1, 'Did not register streamer to hive'
    run_kill_hive(hive)
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Streamer did not yield correct data'


def test_decorated_streamer_with_topic(hive, bee_factory):
    listener = bee_factory.create('listener')
    hive.add(listener)

    @hive.streamer(topic='topic1')
    def stream():
        for i in range(5):
            yield i
    assert len(hive.streamers) == 1, 'Did not register streamer to hive'
    run_kill_hive(hive)
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert listener.calls[0].topic == 'topic1', 'Streamer did not yield correct data'


def test_only_streamers(hive):
    q = Queue()
    d = {'count': 0}

    @hive.streamer
    def publish():
        while True:
            q.put(1)
            yield

    @hive.streamer
    def pull():
        while True:
            try:
                q.get(timeout=0.001)
            except Empty:
                yield
            else:
                d['count'] += 1
                yield

    run_kill_hive(hive)
    hive.close()
    assert d['count'] > 0, 'Did not get any data from queue'


def test_threaded_run(hive, bee_factory):
    stop = _Event()
    listener = bee_factory.create('listener')
    streamer = bee_factory.create('streamer')
    old_on_event = listener.on_event

    def on_event(event):
        if not stop.is_set():
            stop.set()
        else:
            hive.kill()
        return old_on_event(event)

    listener.on_event = on_event
    hive.add(listener)
    hive.add(streamer)
    hive.run(threaded=True)
    stop.wait()
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Streamer did not yield correct data'


def test_close_hive(hive):
    run = Event()

    @hive.streamer
    def stream():
        while True:
            yield 1

    @hive.listener
    def on_event(event):
        run.set()

    def close_after():
        run.wait()
        hive.close()

    Thread(target=close_after).start()
    hive.run()
    assert run.is_set(), 'Did not run on_event before close'


def test_close_hive_only_streamers(hive):
    q = Queue()
    d = {'count': 0}

    @hive.streamer
    def publish():
        while True:
            time.sleep(1e-4)
            q.put(1)
            yield

    @hive.streamer
    def pull():
        while True:
            try:
                q.get(timeout=0.001)
            except Empty:
                yield
            else:
                d['count'] += 1
                yield

    def close_after():
        time.sleep(1e-3)
        hive.close()

    Thread(target=close_after).start()
    hive.run()
    assert d['count'] > 0, 'Did not get any data from queue'


def test_interrupted_run(hive):
    def interrupt():
        time.sleep(5e-3)
        _thread.interrupt_main()

    @hive.streamer
    def stream():
        while True:
            time.sleep(1e-6)
            return 1

    Thread(target=interrupt).start()
    hive.run()
    assert not hive.alive, 'KeyboardInterrupt did not kill hive'


def test_define_with_bad_args(hive):
    with pytest.raises(TypeError):
        hive.streamer(topic={'a': 1})(lambda: None)

    with pytest.raises(TypeError):
        hive.listener(chain=2)(lambda: None)

    with pytest.raises(TypeError):
        hive.listener(filters={'a': 1})(lambda: None)


def test_setup_teardown_exceptions(hive, bee_factory):
    l_failed_setup = bee_factory.create('listener')
    l_failed_teardown = bee_factory.create('listener')
    l_failed_both = bee_factory.create('listener')
    s_failed_setup = bee_factory.create('streamer')
    s_failed_teardown = bee_factory.create('streamer')
    s_failed_both = bee_factory.create('streamer')

    def fail(*args, **kwargs):
        raise ValueError

    for bee in [l_failed_setup, s_failed_setup, l_failed_both, s_failed_both]:
        bee.setup = fail
    for bee in [l_failed_teardown, s_failed_teardown, l_failed_both, s_failed_both]:
        bee.teardown = fail

    for bee in [
        l_failed_setup, s_failed_setup,
        l_failed_teardown, s_failed_teardown,
        l_failed_both, s_failed_both
    ]:
        hive.add(bee)

    run_kill_hive(hive)

    for bee in [l_failed_teardown, l_failed_setup, l_failed_both]:
        assert len(bee.calls) > 0, 'Listener with failed setup did not run'

    for bee in [s_failed_teardown, s_failed_setup, s_failed_both]:
        assert bee.count > 0, 'Streamer with failed setup did not run'

    assert l_failed_setup.teardown_event.is_set(), \
        'Did not attempt to teardown listener with failed setup'
    assert s_failed_setup.teardown_event.is_set(), \
        'Did not attempt to teardown streamer with failed setup'



from threading import Thread, Event as _Event
import asyncio
import time
import _thread

import pybeehive
import pybeehive.async
from pybeehive.async import async_generator
from pybeehive.async.hive import _loop_async


def run_kill_hive(hive, wait=0.005):
    async def kill():
        await asyncio.sleep(wait)
        hive.kill()
    loop = asyncio.get_event_loop()

    hive._set_loop()
    with hive._setup_teardown_listeners():
        with hive._setup_teardown_streamers() as jobs:
            futures = [asyncio.ensure_future(j) for j in jobs]
            loop.run_until_complete(asyncio.gather(
                kill(),
                _loop_async(
                    hive._event_queue, hive.listeners, hive.kill_event
                )
            ))
    for future in futures:
        if not future.done():
            future.cancel()


def test_run(async_hive, async_bee_factory):
    listener = async_bee_factory.create('listener')
    streamer = async_bee_factory.create('streamer')
    async_hive.add(listener)
    async_hive.add(streamer)
    run_kill_hive(async_hive)
    assert len(listener.calls) > 0, 'Run did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Run did not wrap non event data'


def test_decorated_listener(async_hive):
    calls = []

    @async_hive.listener
    async def on_event(event):
        calls.append(event)
        await asyncio.sleep(0)

    assert len(async_hive.listeners) == 1, 'Did not register listener to hive'
    async_hive.submit_event(pybeehive.Event('test'))
    run_kill_hive(async_hive)
    assert len(calls) == 1, 'Listener did not execute on_event'
    assert calls[0].data == 'test', 'Listener did not return correct data'


def test_decorated_streamer(async_hive, async_bee_factory):
    listener = async_bee_factory.create('listener')
    async_hive.add(listener)
    i = 0

    @async_hive.streamer
    @async_generator
    async def stream():
        nonlocal i
        await asyncio.sleep(0)
        i += 1
        if i < 5:
            return i - 1
        else:
            raise StopAsyncIteration
    assert len(async_hive.streamers) == 1, 'Did not register streamer to hive'
    run_kill_hive(async_hive)
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Streamer did not yield correct data'


def test_decorated_streamer_pre_python36(async_hive, async_bee_factory):
    listener = async_bee_factory.create('listener')
    async_hive.add(listener)
    i = 0

    @async_hive.streamer
    @async_generator
    async def stream():
        nonlocal i
        await asyncio.sleep(0)
        i += 1
        if i < 5:
            return i
        raise StopAsyncIteration
    assert len(async_hive.streamers) == 1, 'Did not register streamer to hive'
    run_kill_hive(async_hive)
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Streamer did not yield correct data'


def test_decorated_streamer_pre_python36_closure(async_hive, async_bee_factory):
    listener = async_bee_factory.create('listener')
    async_hive.add(listener)

    @async_hive.streamer
    def stream():
        i = 0

        @async_generator
        async def wrapped():
            nonlocal i
            await asyncio.sleep(0)
            i += 1
            if i < 5:
                return i
            raise StopAsyncIteration
        return wrapped()
    assert len(async_hive.streamers) == 1, 'Did not register streamer to hive'
    run_kill_hive(async_hive)
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Streamer did not yield correct data'


def test_only_streamers(async_hive):
    q = asyncio.Queue()
    d = {'count': 0}

    @async_hive.streamer
    async def publish():
        await q.put(1)
        # need to yield some time to event loop
        # or the test never returns
        await asyncio.sleep(1e-5)

    @async_hive.streamer
    async def pull():
        try:
            q.get_nowait()
        except asyncio.QueueEmpty:
            await asyncio.sleep(1e-4)
        else:
            d['count'] += 1
            await asyncio.sleep(0)

    run_kill_hive(async_hive)
    async_hive.close()
    assert d['count'] > 0, 'Did not get any data from queue'


def test_threaded_run(async_hive, async_bee_factory):
    stop = _Event()
    listener = async_bee_factory.create('listener')
    streamer = async_bee_factory.create('streamer')
    old_on_event = listener.on_event

    async def on_event(event):
        if not stop.is_set():
            stop.set()
        else:
            async_hive.kill()
            await asyncio.sleep(0)
        return await old_on_event(event)

    listener.on_event = on_event
    async_hive.add(listener)
    async_hive.add(streamer)
    async_hive.run(threaded=True)
    stop.wait()
    assert len(listener.calls) > 0, 'Streamer did not yield any events'
    assert isinstance(listener.calls[0], pybeehive.Event), 'Streamer did not yield correct data'


def test_close_hive(async_hive):
    run = _Event()

    @async_hive.streamer
    @async_generator
    async def stream():
        await asyncio.sleep(1e-6)
        return 1

    @async_hive.listener
    async def on_event(event):
        run.set()
        await asyncio.sleep(0)

    def close_after():
        run.wait()
        async_hive.close()

    Thread(target=close_after).start()
    async_hive.run()
    assert run.is_set(), 'Did not run on_event before close'


def test_close_hive_only_streamers(async_hive):
    q = asyncio.Queue()
    d = {'count': 0}

    @async_hive.streamer
    @async_generator
    async def publish():
        await q.put(1)
        # need to yield some time to event loop
        # or the test never returns
        await asyncio.sleep(1e-5)

    @async_hive.streamer
    @async_generator
    async def pull():
        try:
            q.get_nowait()
        except asyncio.QueueEmpty:
            await asyncio.sleep(1e-5)
        else:
            d['count'] += 1
            await asyncio.sleep(0)

    def close_after():
        time.sleep(1e-3)
        async_hive.close()

    Thread(target=close_after).start()
    async_hive.run()
    assert d['count'] > 0, 'Did not get any data from queue'


def test_interrupted_run(async_hive):
    def interrupt():
        time.sleep(5e-3)
        _thread.interrupt_main()

    @async_hive.streamer
    @async_generator
    async def stream():
        await asyncio.sleep(1e-6)
        return 1

    Thread(target=interrupt).start()
    async_hive.run()
    assert not async_hive.alive, 'KeyboardInterrupt did not kill hive'


def test_setup_teardown_exceptions(async_hive, async_bee_factory):
    l_failed_setup = async_bee_factory.create('listener')
    l_failed_teardown = async_bee_factory.create('listener')
    l_failed_both = async_bee_factory.create('listener')
    s_failed_setup = async_bee_factory.create('streamer')
    s_failed_teardown = async_bee_factory.create('streamer')
    s_failed_both = async_bee_factory.create('streamer')

    async def fail(*args, **kwargs):
        raise ValueError

    for bee in [l_failed_setup, s_failed_setup, l_failed_both, s_failed_both]:
        bee.setup = fail
    for bee in [l_failed_teardown, s_failed_teardown, l_failed_both, s_failed_both]:
        bee.teardown = fail

    for bee in [
        l_failed_setup, s_failed_setup,
        l_failed_teardown, s_failed_teardown,
        l_failed_both, s_failed_both
    ]:
        async_hive.add(bee)

    run_kill_hive(async_hive)

    for bee in [l_failed_teardown, l_failed_setup, l_failed_both]:
        assert len(bee.calls) > 0, 'Listener with failed setup did not run'

    for bee in [s_failed_teardown, s_failed_setup, s_failed_both]:
        assert bee.count > 0, 'Streamer with failed setup did not run'

    assert l_failed_setup.teardown_event.is_set(), \
        'Did not attempt to teardown listener with failed setup'
    assert s_failed_setup.teardown_event.is_set(), \
        'Did not attempt to teardown streamer with failed setup'



import time
import pybeehive
from multiprocessing import Queue
import pytest

_now = time.time()
test_event_data = [
        ("test", _now, None),
        ("test", _now, 'topic'),
        ("test\x15", _now, None),
        ("test", _now, 'topic|asraasffsf,\\78*mfjdfb'),
        ("test\r\n\tsdfnskduf*34023849\n\r\\'d", _now, 'abcd'),
        ("", _now, 'abcd'),
        ("abcd", _now, ''),
        ("", _now, None),
        ("a"*100000, _now, None),
        ("test", None, None),
        ("test", None, 'topic'),
        ("test\x15", None, None),
        ("test", None, 'topic|asraasffsf,\\78*mfjdfb'),
        ("test\r\n\tsdfnskduf*34023849\n\r\\'d", None, 'abcd'),
        ("", None, 'abcd'),
        ("abcd", None, ''),
        ("", None, None),
        ("a" * 100000, None, None)
    ]


@pytest.mark.parametrize('data,topic,created_at', test_event_data)
def test_event_string_conversion(data, topic, created_at):
    # for data, created_at, topic in test_event_data:
    event = pybeehive.Event(data, topic=topic, created_at=created_at)
    event_string = event.tostring()
    c_event = pybeehive.Event.fromstring(event_string)
    assert event == c_event, "String converted event not equal to original"


def test_create_event_id():
    now = time.time()
    id_1 = pybeehive.Event.create_id('test', now)
    id_2 = pybeehive.Event.create_id('test', now)
    id_3 = pybeehive.Event.create_id('test', now + 1)
    assert id_1 == id_2, 'Different ids for identical data'
    assert id_1 != id_3 != id_2, 'Same ids for different data'


def test_create_event():
    now = time.time()
    same_1 = pybeehive.Event('test', created_at=now)
    same_2 = pybeehive.Event(same_1)
    assert same_1 == same_2, \
        "Event did not create correctly from another Event"
    diff_1 = pybeehive.Event('test', topic='1', created_at=now)
    diff_2 = pybeehive.Event('test', topic='2', created_at=now)
    diff_3 = pybeehive.Event('test', topic='1', created_at=now + 1)
    diff_4 = pybeehive.Event('test1', topic='1', created_at=now)
    assert diff_1 != diff_2, "Events with different topics equal"
    assert diff_2 != diff_3, "Events with different created_at equal"
    assert diff_3 != diff_4, "Events with different data equal"
    diff_5 = pybeehive.Event(diff_1, topic='2')
    diff_6 = pybeehive.Event(diff_1, created_at=now + 1)
    assert diff_1 != diff_5, "Event created from event with new topic equal"
    assert diff_1 != diff_6, "Event created from event with new time equal"
    assert diff_5.topic == '2', 'Did not override topic correctly'
    assert diff_6.created_at == now + 1, 'Did not override time correctly'


def test_event_dunder_methods():
    now = time.time()
    event_1_a = pybeehive.Event('test_1', created_at=now)
    event_1_b = pybeehive.Event('test_1', created_at=now)
    event_1_c = pybeehive.Event('test_1', created_at=now, topic='thing')
    event_2 = pybeehive.Event('test_1', created_at=now + 1)
    event_3 = pybeehive.Event('test_2', created_at=now)
    assert event_1_a == event_1_b, "Identical events not equal"
    assert event_1_a != event_1_c, "Events with different topics equal"
    assert event_1_a != event_2, "Events with different times equal"
    assert event_1_a != event_3, "Events with different data equal"
    assert str(event_1_a) == str(event_1_b), "Event strings different"
    events = set()
    events.add(event_1_a)
    assert len(events) == 1, "Did not add event to set"
    events.add(event_1_b)
    assert len(events) == 1, "Added identical events to set twice"
    events.add(event_2)
    assert len(events) == 2, "Did not add different event to set"


def test_stream(bee_factory):
    q = Queue()
    streamer = bee_factory.create('streamer')
    streamer.set_queue(q)
    streamer.run()
    time.sleep(0.01)  # because somehow the above line is a race condition ?!
    assert not q.empty(), 'Stream did not yield any events'
    count = 0
    while not q.empty():
        event = q.get()
        assert isinstance(event, pybeehive.Event), 'Did not wrap non event data'
        assert event.data == count, 'Stream did not yield correct data'
        count += 1
    assert count == 10, 'Stream did not yield correct number of events'


def test_stream_with_exception(bee_factory):
    q = Queue()
    streamer = bee_factory.create('streamer', failing=True)
    streamer.set_queue(q)
    streamer.run()
    time.sleep(0.01)  # because somehow the above line is a race condition ?!
    assert not q.empty(), 'Stream did not yield any events'
    assert streamer.ex, 'Stream did not call on_exception'
    count = 0
    while not q.empty():
        event = q.get()
        assert isinstance(event, pybeehive.Event), 'Did not wrap non event data'
        assert event.data == count, 'Stream did not yield correct data'
        count += 1
    assert count == 3, 'Stream did not yield correct number of events'


def test_filter(bee_factory):
    non_filtered = bee_factory.create('listener')
    filtered = bee_factory.create('listener', filters=['topic1'])
    no_topic = pybeehive.Event('data')
    non_filtered.notify(no_topic)
    assert len(non_filtered.calls) == 1, "Listener with no filters did not call on_event"
    filtered.notify(no_topic)
    assert len(filtered.calls) == 0, "Listener with filters called on_event for an event with no topic"
    topic = pybeehive.Event('data', topic='topic1')
    non_filtered.notify(topic)
    assert len(non_filtered.calls) == 2, "Listener with no filters did not call on_event"
    filtered.notify(topic)
    assert len(filtered.calls) == 1, "Listener with filters did not call on_event for an event with topic"


def test_notify(bee_factory):
    listener = bee_factory.create('listener')
    listener.notify(1)
    assert len(listener.calls) == 1, 'on_event did not trigger'
    assert listener.calls[0] == 1, 'Did not process data correct in on_event'
    listener.notify(None)
    assert len(listener.calls) == 1, 'on_event triggered for None'
    assert listener.closed, "Listener did not teardown on kill event"


def test_chained_notify(bee_factory):
    listener1 = bee_factory.create('listener')
    listener2 = bee_factory.create('listener')
    listener3 = bee_factory.create('listener')
    listener1.chain(listener2).chain(listener3)
    listener1.notify(1)
    assert len(listener2.calls) == 1, 'First chained on_event did not trigger'
    assert len(listener3.calls) == 1, 'Second chained on_event did not trigger'
    assert isinstance(listener2.calls[0], pybeehive.Event), 'Did convert data to event in chained on_event'
    assert isinstance(listener3.calls[0], pybeehive.Event), 'Did convert data to event in chained on_event'
    listener1.notify(None)
    assert len(listener2.calls) == 1, 'First chained on_event triggered on None'
    assert len(listener3.calls) == 1, 'Second chained on_event triggered on None'


def test_notify_with_exception(bee_factory):
    listener = bee_factory.create('listener', failing=True)
    listener.notify(1)
    assert len(listener.calls) == 0, 'Failed on_event added data'


def test_chained_notify_with_exception(bee_factory):
    listener1 = bee_factory.create('listener', failing=True)
    listener2 = bee_factory.create('listener', failing=True)
    listener3 = bee_factory.create('listener', failing=True)
    # Ensure both ways of calling work
    pybeehive.Listener.chain([listener1], listener2).chain(listener3)
    listener1.notify(1)
    assert len(listener2.calls) == 0, 'First chained on_event triggered'
    assert len(listener3.calls) == 0, 'Second chained on_event triggered'



import pickle
from abc import ABC, abstractmethod
from threading import Event as _Event
from time import time


class Event:
    """

    :param data:
    :param topic:
    :param created_at:
    """

    def __init__(self, data, topic=None, created_at=None):
        if isinstance(data, Event):
            self.data = data.data
            self.topic = topic or data.topic
            if created_at:
                self.created_at = created_at
                self.id = self.create_id(self.data, created_at)
            else:
                self.created_at = data.created_at
                self.id = data.id
        else:
            self.data = data
            self.topic = topic
            self.created_at = created_at or time()
            self.id = self.create_id(self.data, self.created_at)

    def __eq__(self, other):
        return isinstance(other, Event) \
               and self.id == other.id \
               and self.topic == other.topic

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.id

    def __str__(self):
        return 'Event(created_at={}, data={})'.format(
            int(self.created_at), str(self.data)[:100]
        )

    @staticmethod
    def create_id(data, time_created):
        """

        :param data:
        :param time_created:
        :return:
        """
        return hash(str(data) + str(time_created))

    @staticmethod
    def fromstring(string):
        """

        :param string:
        :return:
        """
        return pickle.loads(string)

    def tostring(self):
        """

        :return:
        """
        return pickle.dumps(self)


class Listener(ABC):
    """

    :param filters:
    """
    def __init__(self, filters=None):
        super(Listener, self).__init__()
        self.chained_bees = []
        self.filters = set(filters or [])

    def __str__(self):
        return "%s(filters=%s)" % (self.__class__.__name__, self.filters)

    def chain(self, bee):
        """

        :param bee:
        :return:
        """
        # static usage with list for many-to-one
        if isinstance(self, list):
            for other in self:
                other.chain(bee)
        else:
            self.chained_bees.append(bee)
        return bee

    def filter(self, event):
        """

        :param event:
        :return:
        """
        # If no filters are defined then listens to all events
        return not self.filters or event.topic in self.filters

    def notify(self, event):
        """

        :param event:
        :return:
        """
        if event and self.filter(event):
            try:
                event = Event(self.on_event(event))
            except Exception as e:
                self.on_exception(e)
            # Only propagate events that this listener can accept
            for bee in self.chained_bees:
                bee.notify(event)

        if not event:
            for bee in self.chained_bees:
                bee.notify(event)
            self.teardown()

    @abstractmethod
    def on_event(self, event):
        """

        :param event:
        :return:
        """
        raise NotImplementedError  # pragma: nocover

    def on_exception(self, exception):
        """

        :param exception:
        """
        pass

    def setup(self):
        """

        """
        pass

    def teardown(self):
        """

        """
        pass


class Killable:
    """

    """
    _event_class = _Event

    def __init__(self):
        self.kill_event = self._event_class()

    @property
    def alive(self):
        """

        :return:
        """
        return not self.kill_event.is_set()

    def kill(self):
        """

        """
        self.kill_event.set()


class Streamer(Killable, ABC):
    """

    :param topic:
    """
    def __init__(self, topic=None):
        super(Streamer, self).__init__()
        self.topic = topic
        self._q = None  # this is set when a streamer is added to a hive

    def __str__(self):
        return "%s(topic=%s)" % (self.__class__.__name__, self.topic)

    @abstractmethod
    def stream(self):
        """

        :return:
        """
        raise NotImplementedError  # pragma: nocover

    def on_exception(self, exception):
        """

        :param exception:
        """
        pass

    def setup(self):
        """

        """
        pass

    def teardown(self):
        """

        """
        pass

    def set_queue(self, q):
        """

        :param q:
        """
        self._q = q

    def run(self):
        """

        """
        self._assert_queue_is_set()
        while self.alive:
            try:
                for data in self.stream():
                    self._q.put(Event(data, topic=self.topic))
                    # break long running streams if the kill event is set
                    if not self.alive:
                        break
            except Exception as e:
                self.on_exception(e)

    def _assert_queue_is_set(self):
        assert self._q is not None, \
            "You must first set the output queue with " \
            "Streamer.set_queue before running a Streamer"



from threading import Thread
import random
import time
import pytest
import _thread

from pybeehive.socket import SocketStreamer, SocketListener
import pybeehive


def test_no_zmq(hive):
    hive._socket_listener_class = None
    hive._socket_streamer_class = None

    with pytest.raises(RuntimeError):
        hive.socket_listener(('', 0))(lambda: None)

    with pytest.raises(RuntimeError):
        hive.socket_streamer(('', 0))(lambda: None)

    hive._socket_listener_class = SocketListener
    hive._socket_streamer_class = SocketStreamer


def test_messaging(client_server):
    msg = b'data'
    client, server = client_server
    server_messages = iter(server.iter_messages())
    for i in range(5):
        client.send(msg)
        received = next(server_messages)
        assert received == msg, 'Incorrect message sent to server'


def test_socket_streamer_listener_decorator_definition(hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    @hive.socket_streamer(address)
    def stream():
        return

    @hive.socket_listener(address)
    def parse_event(event):
        event = pybeehive.Event(event.data + 1, created_at=event.created_at)
        events.append(event)
        return event

    hive.add(SocketStreamer(address))
    hive.submit_event(pybeehive.Event(-1))
    hive.run(threaded=True)
    start = time.time()
    while len(events) < 5 and time.time() - start < 2:
        time.sleep(1e-4)
    hive.close()
    assert len(events) >= 5, "Hive did not process all events"
    for i, e in enumerate(events):
        assert i == e.data, "Event data was not parsed by listener"


def test_socket_streamer_listener_loop(hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    @hive.socket_listener(address)
    def parse_event(event):
        event = pybeehive.Event(event.data + 1, created_at=event.created_at)
        events.append(event)
        return event

    hive.add(SocketStreamer(address))
    hive.submit_event(pybeehive.Event(-1))
    hive.run(threaded=True)
    start = time.time()
    while len(events) < 5 and time.time() - start < 2:
        time.sleep(1e-4)
    hive.close()
    assert len(events) >= 5, "Hive did not process all events"
    for i, e in enumerate(events):
        assert i == e.data, "Event data was not parsed by listener"


def test_multiple_listeners_single_streamer(hive):

    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    class Listener(SocketListener):
        def parse_event(self, event):
            event = pybeehive.Event(event.data + 1, created_at=event.created_at)
            events.append(event)
            return event

    hive.add(SocketStreamer(address))
    for _ in range(3):
        hive.add(Listener(address))

    hive.submit_event(pybeehive.Event(-1))
    hive.run(threaded=True)
    start = time.time()
    while len(events) < 12 and time.time() - start < 2:
        time.sleep(1e-4)
    hive.close()
    assert len(events) >= 12, "Hive did not process all events"
    for i, e in enumerate(events[:12]):
        # First three
        if i < 3:
            assert e.data == 0, "Multiple listeners sent incorrect events"
        # Last nine (each cycle is 3x)
        else:
            assert e.data == 1, "Streamer did not propagate events to listeners"


def test_message_closed_server(hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    @hive.socket_listener(address)
    def parse_event(event):
        events.append(event)
        return event

    hive.submit_event(pybeehive.Event(-1))
    hive.run(threaded=True)
    start = time.time()
    while len(events) < 1 and time.time() - start < 2:
        time.sleep(1e-4)
    hive.close()
    assert len(events) >= 1, "Hive did not process all events"


def test_interrupted_streamer_listener_loop(hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    def interrupt():
        while len(events) < 2:
            time.sleep(1e-3)
        _thread.interrupt_main()

    def parse_event(event):
        events.append(event)
        return event

    streamer = SocketStreamer(address)
    listener = SocketListener(address)
    listener.parse_event = parse_event
    hive.add(streamer)
    hive.add(listener)
    hive.submit_event(pybeehive.Event(-1))
    Thread(target=interrupt).start()
    hive.run()
    assert len(events) > 1, "Listener did not receive any events from streamer"
    assert not hive.alive, 'KeyboardInterrupt did not kill hive'
    assert not streamer.server.alive, 'KeyboardInterrupt did not kill server'
    assert not listener.client.alive, 'KeyboardInterrupt did not kill client'




import inspect
from collections import defaultdict
from contextlib import contextmanager
from queue import Queue, Empty
from threading import Thread

from .core import Listener, Streamer, Event, Killable
from .logging import create_logger, debug_handler, default_handler
try:
    from .socket import SocketListener, SocketStreamer
# This is tested, just not by patching imports
except ImportError:  # pragma: nocover
    SocketListener, SocketStreamer = None, None  # pragma: nocover


def _loop(event_queue, listeners, kill_event):
    while not kill_event.is_set():
        # Timeout on get and Empty catch ensure threads
        # are not waiting forever for an item in the queue
        try:
            event = event_queue.get(timeout=0.001)
            for bee in listeners:
                bee.notify(event)
        except Empty:
            continue
        except KeyboardInterrupt:
            break


class Hive(Killable):
    """

    """
    _listener_class = Listener
    _streamer_class = Streamer
    _socket_listener_class = SocketListener
    _socket_streamer_class = SocketStreamer

    def __init__(self):
        super(Hive, self).__init__()
        self.streamers = []
        self.listeners = _ListenerTree()
        self._event_queue = Queue()

        self.logger = create_logger(handler=default_handler)

    def add(self, *bees):
        """

        :param bees:
        """
        for bee in bees:
            try:
                assert isinstance(bee, Listener), \
                    'Bee must be an instance of Listener or Streamer'
                self.listeners.add_listener(bee)
            except AssertionError:
                assert isinstance(bee, Streamer), \
                    'Bee must be an instance of Listener or Streamer'
                bee.set_queue(self._event_queue)
                self.streamers.append(bee)

    def listener(self, chain=None, filters=None, **kwargs):
        """

        :param chain:
        :param filters:
        :param kwargs:
        :return:
        """
        # for single decorator usage 'chain' is the on_event function
        if inspect.isfunction(chain):
            self._create_listener(chain, **kwargs)
        else:
            if chain:
                self.listeners.validate_chain(chain)

            if filters:
                if not isinstance(filters, (list, set)):
                    filters = [filters]
                self.listeners.validate_filters(filters)

            def wrapper(f):
                self._create_listener(f, chain, filters, **kwargs)
            return wrapper

    def streamer(self, topic=None, **kwargs):
        """

        :param topic:
        :param kwargs:
        :return:
        """
        # for single decorator usage 'topic' is the stream function
        if inspect.isfunction(topic):
            self._create_streamer(topic, **kwargs)
        else:
            if topic:
                self.listeners.validate_filters([topic])

            def wrapper(f):
                self._create_streamer(f, topic=topic, **kwargs)
            return wrapper

    def socket_listener(self, address, chain=None, filters=None):
        """

        :param address:
        :param chain:
        :param filters:
        :return:
        """
        if self._socket_listener_class is None:
            raise RuntimeError('pyzmq required to create pybeehive sockets')

        def wrapped(f):
            return self.listener(
                chain=chain, filters=filters,
                klass=self._socket_listener_class, klass_args=(address,),
                method_name='parse_event'
            )(f)
        return wrapped

    def socket_streamer(self, address, topic=None):
        """

        :param address:
        :param topic:
        :return:
        """
        if self._socket_streamer_class is None:
            raise RuntimeError('pyzmq required to create pybeehive sockets')

        def wrapped(f):
            return self.streamer(
                topic=topic,
                klass=self._socket_streamer_class, klass_args=(address,),
                method_name=None
            )(f)
        return wrapped

    def submit_event(self, event):
        """

        :param event:
        :return:
        """
        assert isinstance(event, Event), "Can only submit Events to the Hive"
        self._event_queue.put_nowait(event)

    def run(self, threaded=False, debug=False):
        """

        :param threaded:
        :param debug:
        :return:
        """
        if debug:
            self.logger.addHandler(debug_handler)
        if threaded:
            worker = Thread(target=self._run)
            worker.start()
            return worker
        else:
            self._run()

    def close(self):
        """

        :return:
        """
        self.kill()
        for streamer in self.streamers:
            streamer.kill()

    def _wrap_stream(self, stream_func):
        return lambda s: stream_func()

    def _create_listener(self, func, chain=None, filters=None,
                         klass=None, klass_args=(), method_name='on_event'):

        _Listener = type(func.__name__, (klass or self._listener_class,), {
            method_name: lambda s, e: func(e)
        })
        self.listeners.add_listener(
            _Listener(*klass_args, filters=filters), chain=chain
        )

    def _create_streamer(self, func, topic=None,
                         klass=None, klass_args=(), method_name='stream'):
        if method_name == 'stream':
            klass_dict = {method_name: self._wrap_stream(func)}
        else:
            klass_dict = {}
        klass = klass or self._streamer_class
        _Streamer = type(func.__name__, (klass,), klass_dict)
        self.add(_Streamer(*klass_args, topic=topic))

    def _run(self):
        with self._setup_teardown_streamers():
            with self._setup_teardown_listeners():
                self.logger.info("The hive is now live!")
                try:
                    _loop(self._event_queue, self.listeners, self.kill_event)
                finally:
                    self.logger.info("Shutting down hive...")
        self.close()

    @contextmanager
    def _setup_teardown_listeners(self):
        self.listeners.call_method_recursively('setup')
        yield
        self.listeners.call_method_recursively('teardown')

    @contextmanager
    def _setup_teardown_streamers(self):
        for streamer in self.streamers:
            self._setup_streamer(streamer)

        yield

        for streamer in self.streamers:
            self._teardown_streamer(streamer)

    def _setup_streamer(self, streamer):
        try:
            streamer.setup()
        except Exception as e:
            self.logger.exception("setup %s - %s", str(streamer), repr(e))
        else:
            self.logger.debug("setup %s - OK", str(streamer))
        streamer.thread = Thread(target=streamer.run)
        streamer.thread.start()

    def _teardown_streamer(self, streamer):
        try:
            streamer.kill()
            streamer.teardown()
            streamer.thread.join()
            self.logger.debug("teardown %s - OK", str(streamer))
        except Exception as e:
            self.logger.exception("teardown %s - %s", str(streamer), repr(e))


class _ListenerTree:
    def __init__(self):
        self._listeners = defaultdict(list)
        self.logger = create_logger(name='pybeehive.hive.listeners')

    def __iter__(self):
        for v in self._listeners.values():
            for l in v:
                yield l

    def __len__(self):
        length = 0
        for _ in self:
            length += 1
        return length

    def add_listener(self, listener, chain=None):
        if chain:
            self.chain(listener, chain)
        else:
            self._listeners[listener.__class__.__name__].append(listener)

    def call_method_recursively(self, method_name, *args, **kwargs):
        results = []
        for bee in self._recursive_iter():
            try:
                result = bee.__getattribute__(method_name)(*args, **kwargs)
            except Exception as e:
                self.logger.exception(
                    "%s %s - %s", method_name, str(bee), repr(e))
            else:
                self.logger.debug("%s %s - OK", method_name, str(bee))
                results.append(result)
        return results

    def chain(self, listener, chain):
        bees_to_chain = []

        if isinstance(chain, str):
            bees_to_chain.extend(self.listeners_by_name(chain))

        else:
            for c in chain:
                bees_to_chain.extend(self.listeners_by_name(c))

        for bee in bees_to_chain:
            bee.chain(listener)

    def listeners_by_name(self, name):
        listeners = []
        for listener in self._recursive_iter():
            if listener.__class__.__name__ == name:
                listeners.append(listener)
        return listeners

    @staticmethod
    def validate_chain(c):

        if isinstance(c, str):
            return

        elif isinstance(c, list) and all(isinstance(i, str) for i in c):
            return

        else:
            raise TypeError("chain must be a string or a list of strings")

    @staticmethod
    def validate_filters(items):
        for item in items:
            try:
                hash(item)
            except TypeError:
                raise TypeError("filter %s is not hashable" % str(item))

    def _recursive_iter(self, listeners=None, visited=None):
        visited = visited or []
        if listeners is None:
            listeners = self

        for listener in listeners:
            if listener not in visited:
                visited.append(listener)
                yield listener
            for _l in self._recursive_iter(listener.chained_bees, visited):
                if _l not in visited:
                    visited.append(_l)
                    yield _l



import asyncio
import pybeehive
import pybeehive.async


def test_stream(async_bee_factory, run_in_loop):
    streamer = async_bee_factory.create('streamer')
    q = asyncio.Queue()
    streamer.set_queue(q)
    run_in_loop(streamer.run)
    assert not q.empty(), 'Stream did not yield any events'
    count = 0
    while not q.empty():
        event = q.get_nowait()
        assert isinstance(event, pybeehive.Event), 'Stream did not wrap non event data'
        assert event.data == count, 'Stream did not yield correct data'
        count += 1
    assert count == 10, 'Stream did not yield correct number of events'


def test_stream_pre_python36(async_bee_factory, run_in_loop):
    streamer = async_bee_factory.create('prepython36')
    results = []

    async def r():
        async for result in streamer.stream():
            results.append(result)

    run_in_loop(r)
    assert len(results) != 0, 'Generator stream did not yield any events'
    count = 0
    for i in results:
        count += 1
        assert i == count, 'Generator stream did not yield correct data'
    assert count == 12, 'Generator stream did not yield correct number of events'


def test_stream_with_exception(async_bee_factory, run_in_loop):
    streamer = async_bee_factory.create('streamer', failing=True)
    q = asyncio.Queue()
    streamer.set_queue(q)
    run_in_loop(streamer.run)
    assert not q.empty(), 'Stream did not yield any events'
    assert streamer.ex, 'Stream did not call on_exception'
    count = 0
    while not q.empty():
        event = q.get_nowait()
        assert isinstance(event, pybeehive.Event), 'Stream did not wrap non event data'
        assert event.data == count, 'Stream did not yield correct data'
        count += 1
    assert count == 3, 'Stream did not yield correct number of events'


def test_notify(async_bee_factory, run_in_loop):
    listener = async_bee_factory.create('listener')
    run_in_loop(listener.notify, 1)
    assert len(listener.calls) == 1, 'on_event did not trigger'
    assert listener.calls[0] == 1, 'Did not process data correct in on_event'


def test_chained_notify(async_bee_factory, run_in_loop):
    listener1 = async_bee_factory.create('listener')
    listener2 = async_bee_factory.create('listener')
    listener3 = async_bee_factory.create('listener')
    listener1.chain(listener2).chain(listener3)
    run_in_loop(listener1.notify, 1)
    assert len(listener2.calls) == 1, 'First chained on_event did not trigger'
    assert len(listener3.calls) == 1, 'Second chained on_event did not trigger'
    assert isinstance(listener2.calls[0], pybeehive.Event), 'Did convert data to event in chained on_event'
    assert isinstance(listener3.calls[0], pybeehive.Event), 'Did convert data to event in chained on_event'


def test_notify_with_exception(async_bee_factory, run_in_loop):
    listener = async_bee_factory.create('listener', failing=True)
    run_in_loop(listener.notify, 1)
    assert len(listener.calls) == 0, 'Failed on_event added data'


def test_chained_notify_with_exception(async_bee_factory, run_in_loop):
    listener1 = async_bee_factory.create('listener', failing=True)
    listener2 = async_bee_factory.create('listener', failing=True)
    listener3 = async_bee_factory.create('listener', failing=True)
    listener1.chain(listener2).chain(listener3)
    run_in_loop(listener1.notify, 1)
    assert len(listener2.calls) == 0, 'First chained on_event triggered'
    assert len(listener3.calls) == 0, 'Second chained on_event triggered'




import asyncio
import random
import pytest
import sys, os
sys.path.append(os.path.abspath('../pybeehive'))
import pybeehive
import pybeehive.async.socket
from pybeehive.async.utils import async_generator


@pytest.fixture
def run_in_loop():
    def wrapped(f, *args, **kwargs):
        return asyncio.get_event_loop().run_until_complete(f(*args, **kwargs))
    return wrapped


@pytest.fixture
def run_in_new_loop(request):
    old_loop = asyncio.get_event_loop()
    loop = asyncio.get_event_loop_policy().new_event_loop()
    asyncio.set_event_loop(loop)

    def wrapped(f, *args, **kwargs):
        result = loop.run_until_complete(f(*args, **kwargs))
        asyncio.set_event_loop(old_loop)
        return result

    return wrapped


# Listener / Streamer
# ===================

class AsyncTestListener(pybeehive.async.Listener):
    def __init__(self):
        super(AsyncTestListener, self).__init__()
        self.calls = []
        self.setup_event = asyncio.Event()
        self.teardown_event = asyncio.Event()

    async def setup(self):
        self.setup_event.set()

    async def teardown(self):
        self.teardown_event.set()

    async def failed_on_event(self, event):
        raise Exception

    async def on_event(self, event):
        self.calls.append(event)
        await asyncio.sleep(0)
        return event


class AsyncTestStreamer(pybeehive.async.Streamer):
    def __init__(self):
        super(AsyncTestStreamer, self).__init__()
        self.count = 0
        self.ex = False
        self.i = 0
        self.setup_event = asyncio.Event()
        self.teardown_event = asyncio.Event()

    async def setup(self):
        self.setup_event.set()

    async def teardown(self):
        self.teardown_event.set()

    @async_generator
    async def stream(self):
        if self.i < 10:
            self.i += 1
            self.count += 1
            await asyncio.sleep(0)
            return self.i - 1
        else:
            self.kill()
            raise StopAsyncIteration

    @async_generator
    async def failed_stream(self):
        if self.i < 3:
            self.i += 1
            await asyncio.sleep(0)
            return self.i - 1
        else:
            self.kill()
            raise Exception

    def on_exception(self, exception):
        self.ex = True


class PrePython36Streamer(pybeehive.async.Streamer):
    def __init__(self):
        super(PrePython36Streamer, self).__init__()
        self.i = 0

    @async_generator
    async def stream(self):
        self.i += 1
        if self.i <= 12:
            return self.i
        else:
            raise StopAsyncIteration


# Hive and bee creation
# =====================

@pytest.fixture
def async_bee_factory(request):
    class BeeFactory(object):
        @staticmethod
        def create(_type, failing=False):
            if _type == 'streamer':
                bee = AsyncTestStreamer()
                if failing:
                    bee.stream = bee.failed_stream
            elif _type == 'listener':
                bee = AsyncTestListener()
                if failing:
                    bee.on_event = bee.failed_on_event
            else:
                bee = PrePython36Streamer()
            return bee
    return BeeFactory()


@pytest.fixture
def async_hive():
    return pybeehive.async.Hive()


# Socket client/server
# ====================

@pytest.fixture
def async_client_server():
    loop = asyncio.get_event_loop()
    port = random.randint(7000, 10000)
    server = pybeehive.async.socket.Server(('127.0.0.1', port))
    client = pybeehive.async.socket.Client(('127.0.0.1', port))
    loop.run_until_complete(server.start())
    loop.run_until_complete(client.connect())
    return client, server





from threading import Thread
import asyncio
import random
import time
import pytest
import _thread

from pybeehive.async.socket import SocketListener, SocketStreamer
import pybeehive


def test_no_zmq(async_hive):
    async_hive._socket_listener_class = None
    async_hive._socket_streamer_class = None

    with pytest.raises(RuntimeError):
        async_hive.socket_listener(('', 0))(lambda: None)

    with pytest.raises(RuntimeError):
        async_hive.socket_streamer(('', 0))(lambda: None)

    async_hive._socket_listener_class = SocketListener
    async_hive._socket_streamer_class = SocketStreamer


# If run_in_new_loop is not the first argument, things break
def test_messaging(run_in_new_loop, async_client_server):
    client, server = async_client_server

    async def _test():
        msg = b'data'
        generator = server.iter_messages().__aiter__()
        for _ in range(5):
            await client.send(msg)
            received = await generator.__anext__()
            assert received == msg, 'Incorrect message sent to server'
        await client.shutdown()
        await server.shutdown()

    run_in_new_loop(_test)


def test_socket_streamer_listener_loop(async_hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    @async_hive.socket_listener(address)
    async def parse_event(event):
        event = pybeehive.Event(event.data + 1, created_at=event.created_at)
        events.append(event)
        return event

    async_hive.add(SocketStreamer(address))
    async_hive.submit_event(pybeehive.Event(-1))
    async_hive.run(threaded=True)
    start = time.time()
    while len(events) < 5 and time.time() - start < 2:
        time.sleep(1e-4)
    async_hive.close()
    assert len(events) >= 5, "Hive did not process all events"
    for i, e in enumerate(events):
        assert i == e.data, "Event data was not parsed by listener"


def test_multiple_listeners_single_streamer(async_hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    class Listener(SocketListener):
        async def parse_event(self, event):
            event = pybeehive.Event(event.data + 1, created_at=event.created_at)
            events.append(event)
            return event

    async_hive.add(SocketStreamer(address))
    for _ in range(3):
        async_hive.add(Listener(address))

    async_hive.submit_event(pybeehive.Event(-1))
    async_hive.run(threaded=True)
    start = time.time()
    while len(events) < 12 and time.time() - start < 2:
        time.sleep(1e-4)
    async_hive.close()
    assert len(events) >= 12, "Hive did not process all events"
    for i, e in enumerate(events[:12]):
        # First three
        if i < 3:
            assert e.data == 0, "Multiple listeners sent incorrect events"
        # Last nine (each cycle is 3x)
        else:
            assert e.data == 1, "Streamer did not propagate events to listeners"


def test_message_closed_server(async_hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    @async_hive.socket_listener(address)
    async def parse_event(event):
        events.append(event)
        return event

    async_hive.submit_event(pybeehive.Event(-1))
    async_hive.run(threaded=True)
    start = time.time()
    while len(events) < 1 and time.time() - start < 2:
        time.sleep(1e-4)
    async_hive.kill()
    assert len(events) >= 1, "Hive did not process all events"


# This test can pass, but due to the inconsistent handling of
# KeyboardInterrupt in zmq.asyncio.Socket.send it does not
# pass consistently and is thus skipped in the normal tests.
# If pyzmq is updated to fix this the skip can be removed.
@pytest.mark.skip
def test_interrupted_streamer_listener_loop(async_hive):
    address = '127.0.0.1', random.randint(7000, 10000)
    events = []

    def interrupt():
        time.sleep(1e-2)
        _thread.interrupt_main()

    async def parse_event(event):
        await asyncio.sleep(1e-4)
        events.append(event)
        return event

    streamer = SocketStreamer(address)
    listener = SocketListener(address)
    listener.parse_event = parse_event
    async_hive.add(streamer)
    async_hive.add(listener)
    async_hive.submit_event(pybeehive.Event(-1))
    Thread(target=interrupt).start()
    async_hive.run()
    assert len(events) > 1, "Listener did not receive any events from streamer"
    assert not async_hive.alive, 'KeyboardInterrupt did not kill hive'
    assert not streamer.server.alive, 'KeyboardInterrupt did not kill server'
    assert not listener.client.alive, 'KeyboardInterrupt did not kill client'



from threading import Event as _Event
import pytest
import random
import sys, os
sys.path.append(os.path.abspath('../pybeehive'))
import pybeehive
import pybeehive.socket


# Listener / Streamer
# ===================

class TestListener(pybeehive.Listener):
    def __init__(self, filters=None):
        super(TestListener, self).__init__(filters=filters)
        self.calls = []
        self.closed = False
        self.setup_event = _Event()
        self.teardown_event = _Event()

    def setup(self):
        self.setup_event.set()

    def teardown(self):
        self.closed = True
        self.teardown_event.set()

    def failed_on_event(self, event):
        raise Exception

    def on_event(self, event):
        self.calls.append(event)
        return event


class TestStreamer(pybeehive.Streamer):
    def __init__(self, topic=None):
        super(TestStreamer, self).__init__(topic=topic)
        self.count = 0
        self.ex = False
        self.setup_event = _Event()
        self.teardown_event = _Event()

    def setup(self):
        self.setup_event.set()

    def teardown(self):
        self.teardown_event.set()

    def stream(self):
        for i in range(10):
            self.count += 1
            yield i
        self.kill()

    def failed_stream(self):
        for i in range(3):
            yield i
        self.kill()
        raise Exception

    def on_exception(self, exception):
        self.ex = True


# Hive and bee creation
# =====================

@pytest.fixture
def bee_factory(request):
    class BeeFactory(object):
        @staticmethod
        def create(_type, failing=False, *args, **kwargs):
            if _type == 'streamer':
                bee = TestStreamer(*args, **kwargs)
                if failing:
                    bee.stream = bee.failed_stream
            else:
                bee = TestListener(*args, **kwargs)
                if failing:
                    bee.on_event = bee.failed_on_event
            request.addfinalizer(bee.teardown)
            return bee
    return BeeFactory()


@pytest.fixture
def hive():
    return pybeehive.Hive()


@pytest.fixture
def client_server(request):
    port = random.randint(7000, 10000)
    server = pybeehive.socket.Server(('127.0.0.1', port))
    client = pybeehive.socket.Client(('127.0.0.1', port))
    server.start()
    client.connect()
    request.addfinalizer(lambda: (client.shutdown(), server.shutdown()))
    return client, server



import logging
import sys

debug_handler = logging.StreamHandler(sys.stderr)
debug_handler.setFormatter(
    logging.Formatter(
        fmt='[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        datefmt='%y-%y-%d %H:%M:%S'
    )
)

default_handler = logging.NullHandler()


def create_logger(name='pybeehive.hive', handler=None):
    logger = logging.getLogger(name)
    if name == 'pybeehive.hive':
        logger.propagate = False
    # set handler only for logger in top level module
    if handler:
        logger.addHandler(handler)
    return logger



from queue import Empty, Queue
from threading import Thread
import time
import zmq
from .core import Streamer, Listener, Event, Killable


class Server(Killable):
    def __init__(self, address):
        super(Server, self).__init__()
        self.address = address
        self.queue = Queue()
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PULL)
        self._listener_thread = None

    def _receive_into_queue(self):
        while self.alive:
            try:
                data = self.socket.recv(flags=zmq.NOBLOCK)
            except zmq.error.ZMQError:
                time.sleep(1e-6)
            else:
                self.queue.put(data)

    def iter_messages(self):
        while not self.kill_event.is_set():
            try:
                msg = self.queue.get(timeout=0.001)
                yield msg
            except Empty:
                continue

    def start(self):
        self.socket.bind("tcp://%s:%s" % self.address)
        self._listener_thread = Thread(target=self._receive_into_queue)
        self._listener_thread.start()

    def shutdown(self):
        self.kill()
        try:
            self._listener_thread.join()
        except AttributeError:
            pass  # there was an error in start, so _listener_thread is None
        self.socket.close(linger=0)


class Client(Killable):
    def __init__(self, address):
        super(Client, self).__init__()
        self.address = address
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUSH)

    def send(self, data):
        while self.alive:
            return self.socket.send(data, flags=zmq.NOBLOCK)

    def connect(self):
        self.socket.connect('tcp://%s:%s' % self.address)

    def shutdown(self):
        self.kill_event.set()
        self.socket.close(linger=0)


class SocketStreamer(Streamer):
    """

    :param address:
    :param topic:
    """
    def __init__(self, address, topic=None):
        super(SocketStreamer, self).__init__(topic=topic)
        self.server = Server(address)

    def setup(self):
        self.server.start()

    def teardown(self):
        self.server.shutdown()

    def stream(self):
        for msg in self.server.iter_messages():
            event = Event.fromstring(msg)
            yield Event(
                event.data, topic=self.topic, created_at=event.created_at
            )


class SocketListener(Listener):
    """

    :param address:
    :param filters:
    """
    def __init__(self, address, filters=None):
        super(SocketListener, self).__init__(filters=filters)
        self.client = Client(address)

    def setup(self):
        self.client.connect()

    def teardown(self):
        self.client.shutdown()

    def on_event(self, event):
        result = self.parse_event(event)
        self.client.send(result.tostring())

    def parse_event(self, event):
        """

        :param event:
        :return:
        """
        return event  # pragma: nocover
